with Ada.Streams;

with P42.Streams;

package body P42.Broadcast.Directory is

   use type Ada.Streams.Stream_Element_Offset;

   --------------------
   -- Directory_Type --
   --------------------

   protected body Directory_Type is

      procedure Publish_Entry;
      procedure Send_Message (Message : in Message_Type);

      procedure Receive_New_Message
        (Data : in Ada.Streams.Stream_Element_Array;
         Address : in GNAT.Sockets.Sock_Addr_Type) is
         Stream : aliased Streams.Streamer_Type (Message_Type'Size / 8);
         pragma Unreferenced (Address);
      begin
         Answer_Seen := True;
         Stream.Initialise;
         Stream.Data (Data'Range) := Data;
         Stream.First := Data'First;
         Stream.Last := Data'Last;
         declare
            Message : constant Message_Type :=
              Message_Type'Input (Stream'Access);
         begin
            case Message.Kind is
               when Demand =>
                  case Message.What is
                     when Publish =>
                        if Message.Category = Local_Entry.Category then
                           Publish_Entry;
                        end if;
                     when Publish_All =>
                        Publish_Entry;
                  end case;
               when Publication =>
                  for An_Entry of Entries loop
                     if An_Entry = Message.Directory_Entry then
                        return;
                     end if;
                  end loop;
                  Entries.Append (Message.Directory_Entry);
            end case;
         end;
      end Receive_New_Message;

      entry Wait_For_Answer when Answer_Seen is
      begin
         null;
      end Wait_For_Answer;

      -----------------
      -- Get_Entries --
      -----------------

      function Get_Entries return Directory_Entry_Vectors.Vector is
      begin
         return Entries;
      end Get_Entries;

      function Get_Entries (Category : in Category_Enum)
                           return Directory_Entry_Vectors.Vector is
         Categorised_Entries : Directory_Entry_Vectors.Vector;
      begin
         for An_Entry of Entries loop
            if An_Entry.Category = Category then
               Categorised_Entries.Append (An_Entry);
            end if;
         end loop;
         return Categorised_Entries;
      end Get_Entries;

      -------------------
      -- Clear_Entries --
      -------------------

      procedure Clear_Entries is
      begin
         Entries.Clear;
      end Clear_Entries;

      ---------------------
      -- Get_Local_Entry --
      ---------------------

      function Get_Local_Entry return Directory_Entry_Type is
      begin
         return Local_Entry.all;
      end Get_Local_Entry;

      procedure Connect is
         use GNAT.Sockets;
      begin
         Create_Socket (Socket, Family_Inet, Socket_Datagram);
         Set_Socket_Option
           (Socket,
            IP_Protocol_For_IP_Level,
            (Multicast_TTL, 1));
         Set_Socket_Option
           (Socket,
            IP_Protocol_For_IP_Level,
            (Multicast_Loop, True));
         Set_Socket_Option
           (Socket,
            IP_Protocol_For_IP_Level,
            (Add_Membership,
             Inet_Addr (Group),
             Any_Inet_Addr));
         Address.Port := Group_Port;
         Address.Addr := Inet_Addr (Group);
      end Connect;

      procedure Send_Message (Message : in Message_Type) is
         Stream : aliased Streams.Streamer_Type (Message_Type'Size / 8);
      begin
         Stream.Initialise;
         Message_Type'Output (Stream'Access, Message);
         Send_Message (Socket, Address, Stream.Data (Stream.First .. Stream.Last));
      end Send_Message;

      procedure Publish_Entry is
         Message : constant Message_Type :=
           (Publication, Local_Entry.Category, Local_Entry.all);
      begin
         Send_Message (Message);
      end Publish_Entry;

      procedure Close is
      begin
         GNAT.Sockets.Close_Socket (Socket);
      end Close;

      procedure Ask_For (Category : in Category_Enum) is
         Message : constant Message_Type :=
           (Demand, Category, Publish);
      begin
         Send_Message (Message);
         Answer_Seen := False;
      end Ask_For;

      procedure Ask_For_All is
         Message : constant Message_Type (Demand, Category_Enum'First) :=
           (Demand, Category_Enum'First, Publish_All);
      begin
         Send_Message (Message);
      end Ask_For_All;

      procedure Set_Listen is
      begin
         Answer_Seen := False;
      end Set_Listen;

   end Directory_Type;

end P42.Broadcast.Directory;
