with P42.Streams;

with Ada.Text_IO;

package body P42.Broadcast.Ressources is

   use type Ada.Streams.Stream_Element_Offset;

   --------------------------
   -- Ressource_Shell_Type --
   --------------------------

   protected body Ressource_Shell_Type is

      procedure Publish_Ressource (Ressource : Ressource_Type);
      procedure Send_Message (Message : in Message_Type);

      procedure Subscribe (Callback : Callback_Type) is
      begin
         Callbacks.Append (Callback);
      end Subscribe;

      procedure Send_Message (Message : in Message_Type) is
         Stream : aliased Streams.Streamer_Type (Message_Type'Size / 8);
      begin
         Stream.Initialise;
         Message_Type'Output (Stream'Access, Message);
         Send_Message (Socket, Address, Stream.Data (Stream.First .. Stream.Last));
      end Send_Message;

      -------------------------
      -- Receive_New_Message --
      -------------------------

      procedure Receive_New_Message
        (Data : in Ada.Streams.Stream_Element_Array;
         Address : in GNAT.Sockets.Sock_Addr_Type)
      is
         Stream : aliased Streams.Streamer_Type (Message_Type'Size / 8);
      begin
         Stream.Initialise;
         Stream.Data (Data'Range) := Data;
         Stream.First := Data'First;
         Stream.Last := Data'Last;
         declare
            Message : constant Message_Type :=
              Message_Type'Input (Stream'Access);
         begin
            Ada.Text_IO.Put_Line ("message received " & Message.Kind'Img);
            case Message.Kind is
               when Demand =>
                  Ada.Text_IO.Put_Line ("message name : " & Message.Name);
                  for Local_Ressource of Local_Ressources loop
                     Ada.Text_IO.Put_Line ("ressource name : " &
                                             Get_Name (Local_Ressource));
                     if Message.Name = Get_Name (Local_Ressource) and
                       Message.Category = Local_Ressource.Category
                     then
                        for Callback of Callbacks loop
                           Callback.all (Local_Ressource, Address);
                        end loop;
                        Publish_Ressource (Local_Ressource);
                     end if;
                  end loop;
               when Publication =>
                  Answer_Seen := True;
                  for Ressource of Local_Ressources loop
                     if Ressource = Message.Ressource then
                        return;
                     end if;
                  end loop;
                  for Ressource of Entries loop
                     if Ressource = Message.Ressource then
                        return;
                     end if;
                  end loop;
                  Entries.Append (Message.Ressource);
            end case;
         end;
      end Receive_New_Message;

      entry Wait_For_Answer when Answer_Seen is
      begin
         null;
      end Wait_For_Answer;

      procedure Set_Listen is
      begin
         Answer_Seen := False;
      end Set_Listen;

      --------------
      --  Append  --
      --------------

      procedure Append (Ressource : Ressource_Type) is
      begin
         Local_Ressources.Append (Ressource);
      end Append;

      procedure Remove (Ressource_Name : String) is
         Cursor : Ressource_Vectors.Cursor;
         use type Ressource_Vectors.Cursor;
      begin
         Cursor := Local_Ressources.First;
         loop
            exit when Cursor = Ressource_Vectors.No_Element;
            if Get_Name (Ressource_Vectors.Element (Cursor)) = Ressource_Name then
               Local_Ressources.Delete (Cursor);
               exit;
            end if;
            Cursor := Ressource_Vectors.Next (Cursor);
         end loop;
      end Remove;

      -------------
      -- Connect --
      -------------

      procedure Connect is
         use GNAT.Sockets;
      begin
         Create_Socket (Socket, Family_Inet, Socket_Datagram);
         Set_Socket_Option
           (Socket,
            IP_Protocol_For_IP_Level,
            (Multicast_TTL, 1));
         Set_Socket_Option
           (Socket,
            IP_Protocol_For_IP_Level,
            (Multicast_Loop, True));
         Set_Socket_Option
           (Socket,
            IP_Protocol_For_IP_Level,
            (Add_Membership,
             Inet_Addr (Group),
             Any_Inet_Addr));
         Address.Port := Group_Port;
         Address.Addr := Inet_Addr (Group);
      end Connect;

      -----------
      -- Close --
      -----------

      procedure Close is
      begin
         GNAT.Sockets.Close_Socket (Socket);
      end Close;

      --------------------
      -- Get_Ressources --
      --------------------

      function Get_Ressources return Ressource_Vectors.Vector is
         use type Ressource_Vectors.Vector;
      begin
         return Local_Ressources & Entries;
      end Get_Ressources;

      -------------------
      -- Get_Ressource --
      -------------------

      function Get_Ressource (Name : in String) return Ressource_Type is
      begin
         for Ressource of Entries loop
            if Get_Name (Ressource) = Name then
               return Ressource;
            end if;
         end loop;
         for Local_Ressource of Local_Ressources loop
            if Get_Name (Local_Ressource) = Name then
               return Local_Ressource;
            end if;
         end loop;
         raise Ressource_Not_Found with Name & " not found";
      end Get_Ressource;

      -----------
      -- Clear --
      -----------

      procedure Clear is
      begin
         Entries.Clear;
      end Clear;

      -------------
      -- Ask_For --
      -------------

      procedure Ask_For (Name : in String;
                         Category : in Category_Enum) is
         Message : constant Message_Type :=
           (Demand, Category, Name'Length, Name);
      begin
         Send_Message (Message);
         Answer_Seen := False;
      end Ask_For;

      procedure Publish_Ressource (Ressource : Ressource_Type) is
         Message : constant Message_Type :=
           (Publication, Ressource.Category, 0, Ressource);
      begin
         Ada.Text_IO.Put_Line ("send message");
         Send_Message (Message);
      end Publish_Ressource;

   end Ressource_Shell_Type;

end P42.Broadcast.Ressources;
