with Ada.Strings.Fixed;
with System;

with GNAT.OS_Lib;

with P42.String_Manipulation;
with P42.Swap_Signed_16;
with P42.Swap_Signed_32;
with P42.Swap_Signed_64;
with P42.Swap_Double;
with P42.Swap_IEEE_Float;

package body P42.Images.Fits is

   Fill_Block : constant String (1 .. 80) := (others => ' ');
   procedure Analyse_Keyword (Keyword : in out Keyword_Header_Type);
   function Get_Row_Address (Data_Row : Data_Row_Type) return System.Address;
   function Get_Factor (Kind : Bitpix_Enum) return Positive;
   procedure Swap_Data_Row (Data_Row : in out Data_Row_Type);
   procedure Remove_First_Keywords (Keywords : in out Keyword_Header_Vectors.Vector);
   procedure Append_First_Keywords (The_Image : in out Image_Type)
     with Pre => Associated_Keyword (The_Image.Keywords.First_Element) not in Simple .. Naxis3;

   function Associated_Keyword (Keyword : Keyword_Header_Type) return Keyword_Enum is
      Reduced_Value : constant String := String_Manipulation.Remove_Spaces (Keyword.Raw (1 .. 8));
   begin
      if Reduced_Value = "END" then
         return End_Header;
      end if;
      return Keyword_Enum'Value (Reduced_Value);
   exception
      when others =>
         return Not_A_Standard_Keyword;
   end Associated_Keyword;

   function Get_Value (Keyword : Keyword_Header_Type) return String is
   begin
      if Keyword.As_Comment then
         return Keyword.Raw (11 .. Keyword.Slash_Position - 1);
      else
         return Keyword.Raw (11 .. 80);
      end if;
   end Get_Value;

   function Get_Integer_Value (Keyword : Keyword_Header_Type) return Integer is
      Reduced_Value : constant String := String_Manipulation.Remove_Spaces (Get_Value (Keyword));
   begin
      return Integer'Value (Reduced_Value);
   end Get_Integer_Value;

   function Get_Bitpix_Value (Keyword : Keyword_Header_Type) return Bitpix_Enum is
      Value : constant String := String_Manipulation.Remove_Spaces (Get_Value (Keyword));
   begin
      if Value = "8" then
         return Byte;
      elsif Value = "16" then
         return Sshort;
      elsif Value = "32" then
         return Slong;
      elsif Value = "64" then
         return Slonglong;
      elsif Value = "-32" then
         return Float;
      elsif Value = "-64" then
         return Double;
      else
         raise Program_Error with """" & Value & """ isn't a bitpix supported string";
      end if;
   end Get_Bitpix_Value;

   function Get_Row_Address (Data_Row : Data_Row_Type) return System.Address is
   begin
      case Data_Row.Kind is
         when Byte =>
            return Data_Row.Bytes (1)'Address;
         when Sshort =>
            return Data_Row.Sshorts (1)'Address;
         when Slong =>
            return Data_Row.Slongs (1)'Address;
         when Slonglong =>
            return Data_Row.Slonglongs (1)'Address;
         when Float =>
            return Data_Row.Floats (1)'Address;
         when Double =>
            return Data_Row.Doubles (1)'Address;
      end case;
   end Get_Row_Address;

   function Get_Factor (Kind : Bitpix_Enum) return Positive is
   begin
      case Kind is
         when Byte =>
            return 1;
         when Sshort =>
            return 2;
         when Slong =>
            return 4;
         when Slonglong =>
            return 8;
         when Float =>
            return 4;
         when Double =>
            return 8;
      end case;
   end Get_Factor;

   procedure Swap_Data_Row (Data_Row : in out Data_Row_Type) is
   begin
      case Data_Row.Kind is
         when Byte =>
            null;
         when Sshort =>
            for A_Sshort of Data_Row.Sshorts loop
               P42.Swap_Signed_16.Swap_Item (A_Sshort);
            end loop;
         when Slong =>
            for A_Slong of Data_Row.Slongs loop
               P42.Swap_Signed_32.Swap_Item (A_Slong);
            end loop;
         when Slonglong =>
            for A_Slonglong of Data_Row.Slonglongs loop
               P42.Swap_Signed_64.Swap_Item (A_Slonglong);
            end loop;
         when Float =>
            for A_Float of Data_Row.Floats loop
               P42.Swap_IEEE_Float.Swap_Item (A_Float);
            end loop;
         when Double =>
            for A_Double of Data_Row.Doubles loop
               P42.Swap_Double.Swap_Item (A_Double);
            end loop;
      end case;
   end Swap_Data_Row;

   function Read_Data_Row (Fd : GNAT.OS_Lib.File_Descriptor;
                           Kind : Bitpix_Enum;
                           Length : Positive) return Data_Row_Type;

   function Read_Image (File_Name : String) return Image_Type is
      Fd : GNAT.OS_Lib.File_Descriptor;
      End_Seen : Boolean := False;
      Nb_Read : Integer;
      use type GNAT.OS_Lib.File_Descriptor;
      Image_To_Return : Image_Type;
   begin
      Fd := GNAT.OS_Lib.Open_Read (File_Name, Fmode => GNAT.OS_Lib.Binary);
      if Fd = GNAT.OS_Lib.Invalid_FD then
         raise Invalid_File with "can't open " & File_Name;
      end if;
      loop
         exit when End_Seen;
         for I in 1 .. 36 loop
            declare
               Keyword : Keyword_Header_Type;
            begin
               Nb_Read := GNAT.OS_Lib.Read (Fd, Keyword.Raw (1)'Address, 80);
               if Nb_Read /= 80 then
                  raise Program_Error with "file not conforment with fits standard" & Nb_Read'Img;
               end if;
               Analyse_Keyword (Keyword);
               if not End_Seen then
                  Image_To_Return.Keywords.Append (Keyword);
                  case Associated_Keyword (Keyword) is
                     when Simple =>
                        if String_Manipulation.Remove_Spaces (Get_Value (Keyword)) /= "T" then
                           raise Invalid_File with File_Name & " isn't conformant with standard";
                        end if;
                     when Bitpix =>
                        Image_To_Return.Bitpix := Get_Bitpix_Value (Keyword);
                     when Naxis =>
                        Image_To_Return.Naxis := Get_Integer_Value (Keyword);
                     when Naxis1 =>
                        Image_To_Return.Naxis1 := Get_Integer_Value (Keyword);
                     when Naxis2 =>
                        Image_To_Return.Naxis2 := Get_Integer_Value (Keyword);
                     when Naxis3 =>
                        Image_To_Return.Naxis3 := Get_Integer_Value (Keyword);
                     when End_Header =>
                        End_Seen := True;
                     when Not_A_Standard_Keyword =>
                        null;
                  end case;
               end if;
            end;
         end loop;
      end loop;
      if Image_To_Return.Naxis /= 0 then
         for I in 1 .. Image_To_Return.Naxis3 loop
            declare
               Data_Rows : Data_Row_Vectors.Vector;
            begin
               for J in 1 .. Image_To_Return.Naxis2 loop
                  Data_Rows.Append (Read_Data_Row (Fd,
                                                   Image_To_Return.Bitpix,
                                                   Image_To_Return.Naxis1));
               end loop;
               Image_To_Return.Datas.Append (Data_Rows);
            end;
         end loop;
      end if;
      GNAT.OS_Lib.Close (Fd);
      return Image_To_Return;
   end Read_Image;

   function Read_Data_Row (Fd : GNAT.OS_Lib.File_Descriptor;
                           Kind : Bitpix_Enum;
                           Length : Positive) return Data_Row_Type is
      Nb_Read : Integer;
      Data_Row : Data_Row_Type (Kind => Kind,
                                Length => Length);
      pragma Warnings (Off, Data_Row);
      Factor : constant Positive := Get_Factor (Kind);
      use type System.Bit_Order;
   begin
      Nb_Read := GNAT.OS_Lib.Read (Fd, Get_Row_Address (Data_Row), Length * Factor);
      if Nb_Read /= Factor * Length then
         raise Invalid_File with Nb_Read'Img & " : is the return value of Read_Data_Row";
      end if;
      if System.Default_Bit_Order = System.Low_Order_First then
         Swap_Data_Row (Data_Row);
      end if;
      return Data_Row;
   end Read_Data_Row;

   procedure Analyse_Keyword (Keyword : in out Keyword_Header_Type) is
      Slash_Position : constant Natural := Ada.Strings.Fixed.Index (Keyword.Raw, "/");
   begin
      Keyword.As_Value := Keyword.Raw (9 .. 10) = "= ";
      Keyword.As_Comment := Slash_Position /= 0;
      Keyword.Slash_Position := Slash_Position;
   end Analyse_Keyword;

   procedure Write (An_Image : Image_Type;
                    File_Name : String) is
      Fd : GNAT.OS_Lib.File_Descriptor;
      Keyword_Counter : Natural := 0;
      Nb_Write : Integer;
      Full_Length : Integer;
   begin
      Fd := GNAT.OS_Lib.Create_File (File_Name, Fmode => GNAT.OS_Lib.Binary);
      for Keyword of An_Image.Keywords loop
         Nb_Write := GNAT.OS_Lib.Write (Fd, Keyword.Raw (1)'Address, 80);
         if Nb_Write /= 80 then
            raise Program_Error with Nb_Write'Img & " : can't write all 80 characters";
         end if;
         Keyword_Counter := Keyword_Counter + 1;
      end loop;
      Keyword_Counter := Keyword_Counter mod 36;
      if Keyword_Counter /= 0 then
         for I in 1 .. 36 - Keyword_Counter loop
            Nb_Write := GNAT.OS_Lib.Write (Fd, Fill_Block (1)'Address, 80);
            if Nb_Write /= 80 then
               raise Program_Error with Nb_Write'Img & " : can't write all 80 fill characters";
            end if;
         end loop;
      end if;
      for Data of An_Image.Datas loop
         for Row of Data loop
            Full_Length := Row.Length * Get_Factor (Row.Kind);
            declare
               Row_To_Write : Data_Row_Type := Row;
               use type System.Bit_Order;
            begin
               if System.Default_Bit_Order = System.Low_Order_First then
                  Swap_Data_Row (Row_To_Write);
               end if;
               Nb_Write := GNAT.OS_Lib.Write (Fd,
                                              Get_Row_Address (Row_To_Write),
                                              Full_Length);
            end;
            if Nb_Write /= Full_Length then
               raise Program_Error with Nb_Write'Img & " : can't write all row data" & Full_Length'Img;
            end if;
         end loop;
      end loop;
      GNAT.OS_Lib.Close (Fd);
   end Write;

   function To_Keyword (Name : String;
                        Is_Value : Boolean := False;
                        Value : String := "";
                        Comment : String := "") return Keyword_Header_Type is
      Keyword : Keyword_Header_Type;
      First_Tail : constant String (1 .. 8 - Name'Length) := (others => ' ');
      Full_Name : constant String (1 .. 8) := Name & First_Tail;
      Operator : constant String (1 .. 2) := (if Is_Value then "= " else "  ");
      Base : constant String := Full_Name & Operator & Value;
      Full_Comment : constant String := (if Comment = "" then "" else "/" & Comment);
      Second_Tail : constant String (1 .. 80 - Base'Length - Full_Comment'Length) := (others => ' ');
      Raw : constant String (1 .. 80) := Base & Full_Comment & Second_Tail;
   begin
      Keyword.Raw := Raw;
      Analyse_Keyword (Keyword);
      return Keyword;
   end To_Keyword;

   function To_Matrix (Rows : Data_Row_Vectors.Vector) return Data_Matrix_Type is
      Matrix : Data_Matrix_Type (Kind => Rows.First_Element.Kind,
                                 Length1 => Rows.First_Element.Length,
                                 Length2 => Positive (Rows.Length));
      Index : Positive := 1;
   begin
      for Row of Rows loop
         case Matrix.Kind is
            when Byte =>
               for I in Row.Bytes'Range loop
                  Matrix.Bytes (I, Index) := Row.Bytes (I);
               end loop;
            when Sshort =>
               for I in Row.Sshorts'Range loop
                  Matrix.Sshorts (I, Index) := Row.Sshorts (I);
               end loop;
            when Slong =>
               for I in Row.Slongs'Range loop
                  Matrix.Slongs (I, Index) := Row.Slongs (I);
               end loop;
            when Slonglong =>
               for I in Row.Slonglongs'Range loop
                  Matrix.Slonglongs (I, Index) := Row.Slonglongs (I);
               end loop;
            when Float =>
               for I in Row.Floats'Range loop
                  Matrix.Floats (I, Index) := Row.Floats (I);
               end loop;
            when Double =>
               for I in Row.Doubles'Range loop
                  Matrix.Doubles (I, Index) := Row.Doubles (I);
               end loop;
         end case;
         Index := Index + 1;
      end loop;
      return Matrix;
   end To_Matrix;

   function To_Rows (Matrix : Data_Matrix_Type) return Data_Row_Vectors.Vector is
      Rows : Data_Row_Vectors.Vector;
   begin
      for J in 1 .. Matrix.Length2 loop
         declare
            Row : Data_Row_Type (Kind => Matrix.Kind,
                                 Length => Matrix.Length1);
         begin
            for I in 1 .. Matrix.Length1 loop
               case Matrix.Kind is
                  when Byte =>
                     Row.Bytes (I) := Matrix.Bytes (I, J);
                  when Sshort =>
                     Row.Sshorts (I) := Matrix.Sshorts (I, J);
                  when Slong =>
                     Row.Slongs (I) := Matrix.Slongs (I, J);
                  when Slonglong =>
                     Row.Slonglongs (I) := Matrix.Slonglongs (I, J);
                  when Float =>
                     Row.Floats (I) := Matrix.Floats (I, J);
                  when Double =>
                     Row.Doubles (I) := Matrix.Doubles (I, J);
               end case;
            end loop;
            Rows.Append (Row);
         end;
      end loop;
      return Rows;
   end To_Rows;

   function "+" (Left, Right : Image_Type) return Image_Type is
      Result : Image_Type;
      Left_Keywords : Keyword_Header_Vectors.Vector := Left.Keywords;
      Right_Keywords : Keyword_Header_Vectors.Vector := Right.Keywords;
      use type Data_Vectors.Vector;
   begin
      Result.Naxis := 3;
      Result.Naxis1 := Left.Naxis1;
      Result.Naxis2 := Left.Naxis2;
      Result.Naxis3 := Left.Naxis3 + Right.Naxis3;
      Remove_First_Keywords (Left_Keywords);
      Left_Keywords.Delete_Last;
      Result.Keywords.Append (To_Keyword (Name => "HISTORY",
                                          Is_Value => False,
                                          Value => "Adding two images"));
      Remove_First_Keywords (Right_Keywords);
      Result.Keywords.Append (Left_Keywords);
      Result.Keywords.Append (Right_Keywords);
      Result.Datas := Left.Datas & Right.Datas;
      Append_First_Keywords (Result);
      return Result;
   end "+";

   procedure Append (Base : in out Image_Type;
                     Additionnal : Image_Type) is
      Add_Keywords : Keyword_Header_Vectors.Vector := Additionnal.Keywords;
   begin
      if Base.Naxis = 2 then
         Base.Naxis := 3;
      end if;
      Base.Naxis3 := Base.Naxis3 + Additionnal.Naxis3;
      Remove_First_Keywords (Base.Keywords);
      Base.Keywords.Delete_Last;
      Base.Keywords.Append (To_Keyword (Name => "HISTORY",
                                        Is_Value => False,
                                        Value => "appending an image"));
      Remove_First_Keywords (Add_Keywords);
      Base.Keywords.Append (Add_Keywords);
      Base.Datas.Append (Additionnal.Datas);
      Append_First_Keywords (Base);
   end Append;

   procedure Append (Base : in out Image_Type;
                     Additionnal : Data_Row_Vectors.Vector;
                     History : String) is
   begin
      if Base.Naxis = 2 then
         Base.Naxis := 3;
      end if;
      Remove_First_Keywords (Base.Keywords);
      Base.Naxis3 := Base.Naxis3 + 1;
      Base.Keywords.Delete_Last;
      Base.Keywords.Append (To_Keyword (Name => "HISTORY",
                                        Is_Value => False,
                                        Value => History));
      Base.Keywords.Append (To_Keyword (Name => "END"));
      Base.Datas.Append (Additionnal);
      Append_First_Keywords (Base);
   end Append;

   procedure Remove_First_Keywords (Keywords : in out Keyword_Header_Vectors.Vector) is
      First_Keyword : constant Keyword_Header_Type := Keywords.First_Element;
   begin
      case Associated_Keyword (First_Keyword) is
         when Simple .. Naxis3 =>
            Keywords.Delete_First;
            Remove_First_Keywords (Keywords);
         when others =>
            null;
      end case;
   end Remove_First_Keywords;

   procedure Append_First_Keywords (The_Image : in out Image_Type) is
      Keywords_To_Add : Keyword_Header_Vectors.Vector;
   begin
      Keywords_To_Add.Append (To_Keyword (Name => "SIMPLE",
                                          Is_Value => True,
                                          Value => "T",
                                          Comment => "File Does Conform To FITS Standard"));
      Keywords_To_Add.Append (To_Keyword (Name => "BITPIX",
                                          Is_Value => True,
                                          Value => To_String (The_Image.Bitpix)));
      Keywords_To_Add.Append (To_Keyword (Name => "NAXIS",
                                          Is_Value => True,
                                          Value => The_Image.Naxis'Img));
      Keywords_To_Add.Append (To_Keyword (Name => "NAXIS1",
                                          Is_Value => True,
                                          Value => The_Image.Naxis1'Img));
      if The_Image.Naxis >= 2 then
         Keywords_To_Add.Append (To_Keyword (Name => "NAXIS2",
                                             Is_Value => True,
                                             Value => The_Image.Naxis2'Img));
      end if;
      if The_Image.Naxis >= 3 then
         Keywords_To_Add.Append (To_Keyword (Name => "NAXIS3",
                                             Is_Value => True,
                                             Value => The_Image.Naxis3'Img));
      end if;
      The_Image.Keywords.Prepend (Keywords_To_Add);
   end Append_First_Keywords;

   function To_String (Value : Bitpix_Enum) return String is
   begin
      case Value is
         when Byte =>
            return "8";
         when Sshort =>
            return "16";
         when Slong =>
            return "32";
         when Float =>
            return "-32";
         when Double =>
            return "-64";
         when Slonglong =>
            return "64";
      end case;
   end To_String;

   function Extract_Images (Image_Base : Image_Type; Pattern : Selection_Pattern) return Image_Type is
      Pattern_Index : Positive := Pattern'First;
      New_Image : Image_Type;
   begin
      New_Image.Naxis := Image_Base.Naxis;
      New_Image.Naxis1 := Image_Base.Naxis1;
      New_Image.Naxis2 := Image_Base.Naxis2;
      New_Image.Naxis3 := Image_Base.Naxis3;
      New_Image.Bitpix := Image_Base.Bitpix;
      New_Image.Keywords := Image_Base.Keywords;
      Remove_First_Keywords (New_Image.Keywords);
      New_Image.Keywords.Delete_Last;
      for Data of Image_Base.Datas loop
         if Pattern (Pattern_Index) then
            New_Image.Datas.Append (Data);
         else
            New_Image.Naxis3 := New_Image.Naxis3 - 1;
         end if;
         Pattern_Index := Pattern_Index + 1;
      end loop;
      New_Image.Keywords.Append (To_Keyword (Name => "HISTORY",
                                             Is_Value => False,
                                             Value => "select images on pattern " &
                                               To_String (Pattern)));
      New_Image.Keywords.Append (To_Keyword (Name => "END"));
      Append_First_Keywords (New_Image);
      return New_Image;
   end Extract_Images;

   procedure Append_Keyword (Image_Base : in out Image_Type;
                             Keyword : Keyword_Header_Type) is
   begin
      Image_Base.Keywords.Delete_Last;
      Image_Base.Keywords.Append (Keyword);
      Image_Base.Keywords.Append (To_Keyword (Name => "END"));
   end Append_Keyword;

end P42.Images.Fits;
