with Ada.Containers.Indefinite_Vectors;
with Ada.Containers.Vectors;

package P42.Images.Fits
  with Preelaborate => True is
   type Keyword_Header_Type is record
      Raw : String (1 .. 80);
      As_Value : Boolean := False;
      Value_Last_Index : Natural := 0;
      As_Comment : Boolean := False;
      Slash_Position : Natural := 0;
   end record;

   type Keyword_Enum is (Simple,
                         Bitpix,
                         Naxis,
                         Naxis1, Naxis2, Naxis3, --  only three dimensions supported for the moment
                         End_Header,
                         Not_A_Standard_Keyword);

   function To_Keyword (Name : String;
                        Is_Value : Boolean := False;
                        Value : String := "";
                        Comment : String := "") return Keyword_Header_Type
     with Pre => Name'Length <= 8 and
                 Value'Length <= 70;

   function Associated_Keyword (Keyword : Keyword_Header_Type) return Keyword_Enum;

   function Get_Value (Keyword : Keyword_Header_Type) return String
     with Pre => Keyword.As_Value,
          Inline => True;
   function Get_Integer_Value (Keyword : Keyword_Header_Type) return Integer
     with Pre => Keyword.As_Value,
          Inline => True;

   package Keyword_Header_Vectors is new Ada.Containers.Vectors (Positive,
                                                                 Keyword_Header_Type,
                                                                 "=");

   subtype Bitpix_Enum is General_Kind_Enum
     with Static_Predicate => Bitpix_Enum in Byte | Sshort | Slong | Float | Double | Slonglong;

   function To_String (Value : Bitpix_Enum) return String;

   function Get_Bitpix_Value (Keyword : Keyword_Header_Type) return Bitpix_Enum
     with Pre => Associated_Keyword (Keyword) = Bitpix and Keyword.As_Value;

   type Data_Row_Type (Kind : Bitpix_Enum; Length : Positive) is record
      case Kind is
         when Byte =>
            Bytes : Byte_Array (1 .. Length);
         when Sshort =>
            Sshorts : Sshort_Array (1 .. Length);
         when Slong =>
            Slongs : Slong_Array (1 .. Length);
         when Slonglong =>
            Slonglongs : Slonglong_Array (1 .. Length);
         when Float =>
            Floats : Float_Array (1 .. Length);
         when Double =>
            Doubles : Double_Array (1 .. Length);
      end case;
   end record;

   package Data_Row_Vectors is new Ada.Containers.Indefinite_Vectors (Positive,
                                                                      Data_Row_Type,
                                                                      "=");

   type Data_Matrix_Type (Kind : Bitpix_Enum; Length1 : Positive; Length2 : Positive) is record
      case Kind is
         when Byte =>
            Bytes : Byte_Matrix (1 .. Length1, 1 .. Length2);
         when Sshort =>
            Sshorts : Sshort_Matrix (1 .. Length1, 1 .. Length2);
         when Slong =>
            Slongs : Slong_Matrix (1 .. Length1, 1 .. Length2);
         when Slonglong =>
            Slonglongs : Slonglong_Matrix (1 .. Length1, 1 .. Length2);
         when Float =>
            Floats : Float_Matrix (1 .. Length1, 1 .. Length2);
         when Double =>
            Doubles : Double_Matrix (1 .. Length1, 1 .. Length2);
      end case;
   end record;

   function To_Matrix (Rows : Data_Row_Vectors.Vector) return Data_Matrix_Type;
   function To_Rows (Matrix : Data_Matrix_Type) return Data_Row_Vectors.Vector;

   package Data_Vectors is new Ada.Containers.Vectors (Positive,
                                                       Data_Row_Vectors.Vector,
                                                       Data_Row_Vectors."=");

   type Image_Type is record
      Keywords : Keyword_Header_Vectors.Vector;
      Naxis : Natural := 0;
      Naxis1 : Natural := 1;
      Naxis2 : Natural := 1;
      Naxis3 : Natural := 1;
      Bitpix : Bitpix_Enum;
      Datas : Data_Vectors.Vector;
   end record;

   function Keyword_List_Closed (Image : Image_Type) return Boolean is
      (Image.Keywords.Last_Element.Raw (1 .. 3) = "END");

   function Are_Compatible (Left, Right : Image_Type) return Boolean is
      (Left.Naxis = Right.Naxis and
       Left.Naxis1 = Right.Naxis1 and
       Left.Naxis2 = Right.Naxis2 and
       Left.Bitpix = Right.Bitpix);

   function Read_Image (File_Name : String) return Image_Type
     with Post => Keyword_List_Closed (Read_Image'Result);

   procedure Write (An_Image : Image_Type;
                    File_Name : String);

   function "+" (Left, Right : Image_Type) return Image_Type
     with Pre => Keyword_List_Closed (Left) and
                 Keyword_List_Closed (Right) and
                 Are_Compatible (Left, Right),
          Post => Keyword_List_Closed ("+"'Result);
   procedure Append (Base : in out Image_Type;
                     Additionnal : Image_Type)
     with Pre => Keyword_List_Closed (Base) and
                 Keyword_List_Closed (Additionnal) and
                 Are_Compatible (Base, Additionnal),
          Post => Keyword_List_Closed (Base);
   procedure Append (Base : in out Image_Type;
                     Additionnal : Data_Row_Vectors.Vector;
                     History : String)
     with Pre => Keyword_List_Closed (Base),
          Post => Keyword_List_Closed (Base);

   function Extract_Images (Image_Base : Image_Type; Pattern : Selection_Pattern) return Image_Type
     with Pre => Image_Base.Naxis3  = Pattern'Length and
                 Pattern'Length >= 2 and
                 (for some Pattern_Value of Pattern => Pattern_Value),
          Post => Keyword_List_Closed (Extract_Images'Result);

   procedure Append_Keyword (Image_Base : in out Image_Type;
                             Keyword : Keyword_Header_Type)
     with Post => Keyword_List_Closed (Image_Base);

end P42.Images.Fits;
