--  with Ada.Text_IO;
with System;
with System.Storage_Elements;

with P42.Swap_Unsigned_16;
with P42.Swap_Unsigned_32;
with P42.Swap_Signed_32;
with P42.Swap_IEEE_Float;

package body P42.Images.Tiff is

   function Local_Lseek (FD     : GNAT.OS_Lib.File_Descriptor;
                         Offset : Long_Integer := 0;
                         Origin : Integer := GNAT.OS_Lib.Seek_Cur) return Long_Integer;
   pragma Import (C, Local_Lseek, "lseek");


   -----------
   -- Image --
   -----------

   function Image (File_Name : String; Memory_Footprint_Limit : Long_Integer := One_Giga_Byte) return Image_Type is
      Fd : GNAT.OS_Lib.File_Descriptor;
      Nb_Read : Integer;
      use type GNAT.OS_Lib.File_Descriptor;
      use type System.Bit_Order;
      use type Interfaces.Unsigned_16;
      use type Interfaces.Unsigned_32;
      Need_Swap : Boolean := False;
      Nb_Directory_Entries : Interfaces.Unsigned_16;
      Raw_Directory_Entry : Raw_Directory_Entry_Type;
      Next_IFD_Offset : Interfaces.Unsigned_32 := 0;
      File_Length : Long_Integer;
      Image_To_Return : Image_Type;
      Sample_Format_Present : Boolean := False;
      Bits_Per_Sample_Present : Boolean := False;
   begin
      Fd := GNAT.OS_Lib.Open_Read (File_Name, Fmode => GNAT.OS_Lib.Binary);
      if Fd = GNAT.OS_Lib.Invalid_FD then
         raise Invalid_File with "can't open " & File_Name;
      end if;
      Nb_Read := GNAT.OS_Lib.Read (Fd, Image_To_Return.Header.Endianness'Address, Header_Size);
      if Nb_Read /= Header_Size then
         raise Invalid_File with "can't read first bytes" & Nb_Read'Img;
      end if;
      Need_Swap := (Image_To_Return.Header.Endianness = Big_Endian and
                      System.Default_Bit_Order = System.Low_Order_First) or
        (Image_To_Return.Header.Endianness = Little_Endian and
           System.Default_Bit_Order = System.High_Order_First);
      if Need_Swap then
         Swap_Unsigned_16.Swap_Item (Image_To_Return.Header.Fourty_Two);
         Swap_Unsigned_32.Swap_Item (Image_To_Return.Header.IFD_Offset);
      end if;
      if Image_To_Return.Header.Fourty_Two /= 42 then
         raise Invalid_File with "fourty value :" & Image_To_Return.Header.Fourty_Two'Img;
      end if;
      GNAT.OS_Lib.Lseek (FD => Fd,
                         offset => Long_Integer (Image_To_Return.Header.IFD_Offset),
                         origin => GNAT.OS_Lib.Seek_Set);
      loop
         Nb_Read := GNAT.OS_Lib.Read (Fd, Nb_Directory_Entries'Address, 2);
         if Nb_Read /= 2 then
            raise Invalid_File with "can't read nb of directories" & Nb_Read'Img;
         end if;
         if Need_Swap then
            Swap_Unsigned_16.Swap_Item (Nb_Directory_Entries);
         end if;
         for I in 1 .. Nb_Directory_Entries loop
            Nb_Read := GNAT.OS_Lib.Read (Fd, Raw_Directory_Entry'Address, Directory_Entry_Type_Size);
            if Nb_Read /= Directory_Entry_Type_Size then
               raise Invalid_File with "can't read directory entry" & I'Img & Nb_Read'Img;
            end if;
            if Need_Swap then
               Sample_Format_Present := Sample_Format_Present or
                 Sample_Format = Swap_Tag.Swap_Item (Raw_Directory_Entry.Tag);
               Bits_Per_Sample_Present := Bits_Per_Sample_Present or
                 Bits_Per_Sample = Swap_Tag.Swap_Item (Raw_Directory_Entry.Tag);
            else
               Sample_Format_Present := Sample_Format = Raw_Directory_Entry.Tag;
               Bits_Per_Sample_Present := Bits_Per_Sample = Raw_Directory_Entry.Tag;
            end if;
            Image_To_Return.Directories.Append (To_Directory_Entry (Fd, Raw_Directory_Entry, Need_Swap));
         end loop;
         Nb_Read := GNAT.OS_Lib.Read (Fd, Next_IFD_Offset'Address, 4);
         if Nb_Read /= 4 then
            raise Invalid_File with "can't read next IFD Offset" & Nb_Read'Img;
         end if;
         if Need_Swap then
            Swap_Unsigned_32.Swap_Item (Next_IFD_Offset);
         end if;
         exit when Next_IFD_Offset = 0;
         GNAT.OS_Lib.Lseek (FD => Fd,
                            offset => Long_Integer (Next_IFD_Offset),
                            origin => GNAT.OS_Lib.Seek_Set);
      end loop;
      File_Length := GNAT.OS_Lib.File_Length (Fd);
      if File_Length > Memory_Footprint_Limit then
         raise Unsupported_Feature with File_Length'Img & " is higher than" & Memory_Footprint_Limit'Img;
      end if;
      declare
         Sample_Format_Value : Sample_Format_Enum := Unsigned_Integer;
         Directory_Image_Width : constant Directory_Entry_Type := Find (Image_To_Return.Directories, Image_Width);
         Directory_Image_Length : constant Directory_Entry_Type := Find (Image_To_Return.Directories, Image_Length);
         Directory_Rows_Per_Strip : constant Directory_Entry_Type := Find (Image_To_Return.Directories, Rows_Per_Strip);
         Nb_Strip : Interfaces.Unsigned_32;
         Data_Width : Data_Width_Enum := Width_8;
         Directory_Strip_Offsets : constant Directory_Entry_Type := Find (Image_To_Return.Directories, Strip_Offsets);
         Length : Integer;
         Directory_Photometric_Interpretation : constant Directory_Entry_Type :=
           Find (Image_To_Return.Directories, Photometric_Interpretation);
      begin
         if Sample_Format_Present then
            declare
               Directory_Sample_Format : constant Directory_Entry_Type :=
                 Find (Image_To_Return.Directories, Sample_Format);
            begin
               if Directory_Sample_Format.Shorts.First_Element = 1 then
                  Sample_Format_Value := Unsigned_Integer;
               elsif Directory_Sample_Format.Shorts.First_Element = 2 then
                  Sample_Format_Value := Signed_Integer;
               elsif Directory_Sample_Format.Shorts.First_Element = 3 then
                  Sample_Format_Value := Ieee_Float;
               elsif Directory_Sample_Format.Shorts.First_Element = 4 then
                  raise Unknown_Sample_Format;
               else
                  raise Unknown_Sample_Format with
                    Directory_Sample_Format.Shorts.First_Element'Img & " is an unexpected value";
               end if;
            end;
         else
            Sample_Format_Value := Unsigned_Integer;
         end if;
         if Bits_Per_Sample_Present then
            declare
               Directory_Bits_Per_Sample : constant Directory_Entry_Type :=
                 Find (Image_To_Return.Directories, Bits_Per_Sample);
            begin
               if Directory_Bits_Per_Sample.Shorts.First_Element = 8 then
                  Data_Width := Width_8;
               elsif Directory_Bits_Per_Sample.Shorts.First_Element = 32 then
                  Data_Width := Width_32;
               else
                  raise Unsupported_Data_Width with
                    Directory_Bits_Per_Sample.Shorts.First_Element'Img & " isn't a bits per sample supported yet";
               end if;
            end;
         end if;
         Nb_Strip := Directory_Image_Length.A_Long /
           Interfaces.Unsigned_32 (Directory_Rows_Per_Strip.Shorts.First_Element);
         if Directory_Image_Length.A_Long mod
           Interfaces.Unsigned_32 (Directory_Rows_Per_Strip.Shorts.First_Element) /= 0
         then
            Nb_Strip := Nb_Strip + 1;
         end if;
         if Nb_Strip /= 1 then
            raise Unsupported_Feature with "more than one strip isn't supported yet";
         end if;
         GNAT.OS_Lib.Lseek (FD => Fd,
                            offset => Long_Integer (Directory_Strip_Offsets.A_Long),
                            origin => GNAT.OS_Lib.Seek_Set);
         if Directory_Photometric_Interpretation.Shorts.First_Element <= 1 then
            if Bits_Per_Sample_Present then
               Image_To_Return.Kind := Grayscale;
            else
               Image_To_Return.Kind := Bilevel;
            end if;
         elsif Directory_Photometric_Interpretation.Shorts.First_Element = 2 then
            Image_To_Return.Kind := RGB_Full_Color;
         elsif Directory_Photometric_Interpretation.Shorts.First_Element = 3 then
            Image_To_Return.Kind := Palette_Color;
         else
            raise Unsupported_Feature with
              Directory_Photometric_Interpretation.Shorts.First_Element'Img & " isn't a supported value yet";
         end if;
         if Image_To_Return.Kind = RGB_Full_Color then
            declare
               Directory_Samples_Per_Pixel : constant Directory_Entry_Type :=
                 Find (Image_To_Return.Directories, Samples_Per_Pixel);
            begin
               Length := Positive (Directory_Image_Width.A_Long) *
                 Positive (Directory_Samples_Per_Pixel.Shorts.First_Element);
            end;
         else
            Length := Positive (Directory_Image_Width.A_Long);
         end if;
         for I in 1 .. Directory_Image_Length.A_Long loop
            declare
               Row : Row_Data_Type (Sample_Format_Value, Length, Data_Width);
            begin
               case Sample_Format_Value is
                  when Unsigned_Integer =>
                     case Row.Data_Width is
                        when Width_8 =>
                           Nb_Read := GNAT.OS_Lib.Read (Fd,
                                                        Row.Unsigned_8_Data (1)'Address,
                                                        Row.Unsigned_8_Data'Length);
                           if Nb_Read /= Row.Unsigned_8_Data'Length then
                              raise Invalid_File with "Can't read row data" & Nb_Read'Img;
                           end if;
                        when Width_32 =>
                           Nb_Read := GNAT.OS_Lib.Read (Fd,
                                                        Row.Unsigned_32_Data (1)'Address,
                                                        4 * Row.Unsigned_32_Data'Length);
                           if Nb_Read /= (4 * Row.Unsigned_32_Data'Length) then
                              raise Invalid_File with "Can't read row data" & Nb_Read'Img;
                           end if;
                           if Need_Swap then
                              for I in Row.Unsigned_32_Data'Range loop
                                 Swap_Unsigned_32.Swap_Item (Row.Unsigned_32_Data (I));
                              end loop;
                           end if;
                     end case;
                  when Signed_Integer =>
                     case Row.Data_Width is
                        when Width_8 =>
                           raise Unsupported_Feature with "can't handle 8 bits signed data yet";
                        when Width_32 =>
                           Nb_Read := GNAT.OS_Lib.Read (Fd,
                                                        Row.Signed_32_Data (1)'Address,
                                                        4 * Row.Signed_32_Data'Length);
                           if Nb_Read /= (4 * Row.Signed_32_Data'Length) then
                              raise Invalid_File with "Can't read row data" & Nb_Read'Img;
                           end if;
                           if Need_Swap then
                              for I in Row.Signed_32_Data'Range loop
                                 Swap_Signed_32.Swap_Item (Row.Signed_32_Data (I));
                              end loop;
                           end if;
                     end case;
                  when Ieee_Float =>
                     case Row.Data_Width is
                        when Width_8 =>
                           raise Unsupported_Feature with "can't handle float other that 32 bits";
                        when Width_32 =>
                           Nb_Read := GNAT.OS_Lib.Read (Fd,
                                                        Row.Ieee_Float_Data (1)'Address,
                                                        4 * Row.Ieee_Float_Data'Length);
                           if Nb_Read /= (4 * Row.Ieee_Float_Data'Length) then
                              raise Invalid_File with "Can't read row data" & Nb_Read'Img;
                           end if;
                           if Need_Swap then
                              for I in Row.Ieee_Float_Data'Range loop
                                 Swap_IEEE_Float.Swap_Item (Row.Ieee_Float_Data (I));
                              end loop;
                           end if;
                     end case;
               end case;
               Image_To_Return.Rows.Append (Row);
            end;
         end loop;
      end;
      return Image_To_Return;
   end Image;

   function To_Directory_Entry (Fd : GNAT.OS_Lib.File_Descriptor;
                                Raw : Raw_Directory_Entry_Type;
                                Need_Swap : Boolean) return Directory_Entry_Type is
      Kind : Kind_Enum;
      Length : Natural;
      Count : Interfaces.Unsigned_32;
      Tag : Tag_Enum;
      Offset : Long_Integer;
      Old_Offset : Long_Integer;
      use type Interfaces.Unsigned_32;
      Nb_Read : Integer;
   begin
      if Need_Swap then
         Kind := Swap_Kind_Enum.Swap_Item (Raw.Kind);
         Length := Natural (Swap_Unsigned_32.Swap_Item (Raw.Count));
         Count := Swap_Unsigned_32.Swap_Item (Raw.Count);
         Tag := Swap_Tag.Swap_Item (Raw.Tag);
         Offset := Long_Integer (Swap_Unsigned_32.Swap_Item (Raw.Offset));
      else
         Kind := Raw.Kind;
         Length := Natural (Raw.Count);
         Count := Raw.Count;
         Offset := Long_Integer (Raw.Offset);
         Tag := Raw.Tag;
      end if;
      declare
         Return_Item : Directory_Entry_Type (Kind => Kind);
         A_String : String (1 .. Length);
         Last_Nul_Index : Positive := 1;
      begin
         Return_Item.Sub_Directory := (Raw => Raw,
                                       Tag => Tag,
                                       Count => Count);
         case Kind is
            when Byte =>
               if Count /= 1 then
                  raise Unsupported_Kind with "more than one byte isn't supported yet";
               end if;
               declare
                  C : Interfaces.Unsigned_8;
                  for C'Address use Raw.Offset'Address;
               begin
                  Return_Item.C := C;
               end;
            when Ascii =>
               if Count <= 4 then
                  raise Unsupported_Kind with "small ascii handling isn't supported yet";
               end if;
               Old_Offset := Local_Lseek (Fd);
               GNAT.OS_Lib.Lseek (Fd, Offset, GNAT.OS_Lib.Seek_Set);
               Nb_Read := GNAT.OS_Lib.Read (Fd, A_String (1)'Address, Length);
               GNAT.OS_Lib.Lseek (Fd, Old_Offset, GNAT.OS_Lib.Seek_Set);
               if Nb_Read /= Length then
                  raise Invalid_File with "can't read string" & Nb_Read'Img;
               end if;
               for I in A_String'Range loop
                  if A_String (I) = Standard.Ascii.NUL or A_String (I) = Standard.Ascii.LF then
                     if Last_Nul_Index < I - 1 then
                        Return_Item.Strings.Append (A_String (Last_Nul_Index .. I - 1));
                     end if;
                     Last_Nul_Index := I + 1;
                  end if;
               end loop;
            when Short =>
               if Count = 1 then
                  declare
                     Value_16 : Interfaces.Unsigned_16;
                     for Value_16'Address use Raw.Offset'Address;
                  begin
                     if Need_Swap then
                        Return_Item.Shorts.Append (Swap_Unsigned_16.Swap_Item (Value_16));
                     else
                        Return_Item.Shorts.Append (Value_16);
                     end if;
                  end;
               elsif Count = 2 then
                  declare
                     Value_16_1 : Interfaces.Unsigned_16;
                     for Value_16_1'Address use Raw.Offset'Address;
                     use type System.Storage_Elements.Storage_Offset;
                     Value_16_2 : Interfaces.Unsigned_16;
                     for Value_16_2'Address use Raw.Offset'Address + 2;
                  begin
                     if Need_Swap then
                        Return_Item.Shorts.Append (Swap_Unsigned_16.Swap_Item (Value_16_1));
                        Return_Item.Shorts.Append (Swap_Unsigned_16.Swap_Item (Value_16_2));
                     else
                        Return_Item.Shorts.Append (Value_16_1);
                        Return_Item.Shorts.Append (Value_16_2);
                     end if;
                  end;
               elsif Count >= 3 then
                  Old_Offset := Local_Lseek (Fd);
                  GNAT.OS_Lib.Lseek (Fd, Offset, GNAT.OS_Lib.Seek_Set);
                  for I in 1 .. Count loop
                     declare
                        Value_16 : Interfaces.Unsigned_16;
                     begin
                        Nb_Read := GNAT.OS_Lib.Read (Fd, Value_16'Address, 2);
                        if Nb_Read /= 2 then
                           raise Invalid_File with Tag'Img & " can't read short" & Nb_Read'Img;
                        end if;
                        if Need_Swap then
                           Return_Item.Shorts.Append (Swap_Unsigned_16.Swap_Item (Value_16));
                        else
                           Return_Item.Shorts.Append (Value_16);
                        end if;
                     end;
                  end loop;
                  GNAT.OS_Lib.Lseek (Fd, Old_Offset, GNAT.OS_Lib.Seek_Set);
               end if;
            when Long =>
               if Count /= 1 then
                  raise Unsupported_Kind with "more than one long isn't supported yet";
               end if;
               declare
                  Value_32 : Interfaces.Unsigned_32;
                  for Value_32'Address use Raw.Offset'Address;
               begin
                  if Need_Swap then
                     Return_Item.A_Long := Swap_Unsigned_32.Swap_Item (Value_32);
                  else
                     Return_Item.A_Long := Value_32;
                  end if;
               end;
            when Rational | Srational | Undefined | Double =>
               raise Unsupported_Kind with Kind'Img & " isn't supported yet";
            when Sbyte | Sshort | Slong | Float =>
               raise Unsupported_Kind with Kind'Img & " isn't supported yet";
         end case;
         return Return_Item;
      end;
   end To_Directory_Entry;

   function Find (Directory_Entries : Directory_Entry_Vectors.Vector;
                  Tag : Tag_Enum) return Directory_Entry_Type is
   begin
      for Directory_Entry of Directory_Entries loop
         if Directory_Entry.Sub_Directory.Tag = Tag then
            return Directory_Entry;
         end if;
      end loop;
      raise Tag_Not_Found with Tag'Img & " can't be found";
   end Find;

end P42.Images.Tiff;
