with Ada.Containers.Indefinite_Vectors;
with Ada.Containers.Vectors;
with Interfaces;

with GNAT.OS_Lib;

with P42.String_Vectors;
with P42.Swap;

package P42.Images.Tiff is
   pragma Preelaborate;

   type Image_Endianess_Enum is (Little_Endian, Big_Endian);
   for Image_Endianess_Enum'Size use 16;
   for Image_Endianess_Enum use
     (Little_Endian => 16#49_49#,  --  II
      Big_Endian    => 16#4D_4D#); --  MM

   Header_Size : constant := 8;
   type Image_Header_Type is record
      Endianness : Image_Endianess_Enum;
      Fourty_Two : Interfaces.Unsigned_16;
      IFD_Offset : Interfaces.Unsigned_32;
   end record;
   for Image_Header_Type'Size use Header_Size * 8;
   for Image_Header_Type use record
      Endianness at 0 range 0 .. 15;
      Fourty_Two at 2 range 0 .. 15;
      IFD_Offset at 4 range 0 .. 31;
   end record;

   subtype Kind_Enum is General_Kind_Enum
     with Static_Predicate => Kind_Enum not in Slonglong;

   package Swap_Kind_Enum is new Swap (General_Kind_Enum);

   type Tag_Enum is (New_Subfile_Type,
                     Subfile_Type,
                     Image_Width,
                     Image_Length,
                     Bits_Per_Sample,
                     Compression,
                     Photometric_Interpretation,
                     Threshholding,
                     Cell_Width,
                     Cell_Length,
                     Fill_Order,
                     Image_Description,
                     Make,
                     Model,
                     Strip_Offsets,
                     Orientation,
                     Samples_Per_Pixel,
                     Rows_Per_Strip,
                     Strip_Byte_Counts,
                     Min_Sample_Value,
                     Max_Sample_Value,
                     X_Resolution,
                     Y_Resolution,
                     Planar_Configuration,
                     Free_Offsets,
                     Free_Byte_Counts,
                     Gray_Response_Unit,
                     Gray_Response_Curve,
                     Resolution_Unit,
                     Software,
                     Date_Time,
                     Artist,
                     Host_Computer,
                     Color_Map,
                     Extra_Samples,
                     Sample_Format,
                     S_Min_Sample_Value,
                     S_Max_Sample_Value,
                     Copyright);
   for Tag_Enum use (New_Subfile_Type => 254,
                     Subfile_Type => 255,
                     Image_Width => 256,
                     Image_Length => 257,
                     Bits_Per_Sample => 258,
                     Compression => 259,
                     Photometric_Interpretation => 262,
                     Threshholding => 263,
                     Cell_Width => 264,
                     Cell_Length => 265,
                     Fill_Order => 266,
                     Image_Description => 270,
                     Make => 271,
                     Model => 272,
                     Strip_Offsets => 273,
                     Orientation => 274,
                     Samples_Per_Pixel => 277,
                     Rows_Per_Strip => 278,
                     Strip_Byte_Counts => 279,
                     Min_Sample_Value => 280,
                     Max_Sample_Value => 281,
                     X_Resolution => 282,
                     Y_Resolution => 283,
                     Planar_Configuration => 284,
                     Free_Offsets => 288,
                     Free_Byte_Counts => 289,
                     Gray_Response_Unit => 290,
                     Gray_Response_Curve => 291,
                     Resolution_Unit => 296,
                     Software => 305,
                     Date_Time => 306,
                     Artist => 315,
                     Host_Computer => 316,
                     Color_Map => 320,
                     Extra_Samples => 338,
                     Sample_Format => 339,
                     S_Min_Sample_Value => 340,
                     S_Max_Sample_Value => 341,
                     Copyright => 33432);
   for Tag_Enum'Size use 16;

   package Swap_Tag is new Swap (Tag_Enum);

   Directory_Entry_Type_Size : constant := 12;
   type Raw_Directory_Entry_Type is record
      Tag : Tag_Enum;
      Kind : Kind_Enum;  --  TIFF6 specifications use type instead of kind
      Count : Interfaces.Unsigned_32;
      Offset : Interfaces.Unsigned_32;
   end record;
   for Raw_Directory_Entry_Type'Size use Directory_Entry_Type_Size * 8;

   type Sub_Directory_Type is record
      Raw : Raw_Directory_Entry_Type;
      Count : Interfaces.Unsigned_32;
      Tag : Tag_Enum;
   end record;

   package Short_Vectors is new Ada.Containers.Vectors (Positive,
                                                        Interfaces.Unsigned_16,
                                                        Interfaces."=");

   type Directory_Entry_Type (Kind : Kind_Enum) is record
      Sub_Directory : Sub_Directory_Type;
      case Kind is
         when Byte =>
            C : Interfaces.Unsigned_8;
         when Ascii =>
            Strings : String_Vectors.Vector;
         when Short =>
            Shorts : Short_Vectors.Vector;
         when Long =>
            A_Long : Interfaces.Unsigned_32;
         when Rational =>
            Numerator : Interfaces.Unsigned_32;
            Denominator : Interfaces.Unsigned_32;
         when Sbyte =>
            S_C : Interfaces.Integer_8;
         when Undefined =>
            null;
         when Sshort =>
            S_Short : Interfaces.Integer_16;
         when Slong =>
            S_Long : Interfaces.Integer_32;
         when Srational =>
            S_Numerator : Interfaces.Unsigned_32;
            S_Denominator : Interfaces.Unsigned_32;
         when Float =>
            A_Float : Standard.Float;
         when Double =>
            A_Double : Long_Float;
      end case;
   end record;

   Unsupported_Kind : exception;

   function To_Directory_Entry (Fd : GNAT.OS_Lib.File_Descriptor;
                                Raw : Raw_Directory_Entry_Type;
                                Need_Swap : Boolean) return Directory_Entry_Type;

   package Directory_Entry_Vectors is new Ada.Containers.Indefinite_Vectors (Positive,
                                                                             Directory_Entry_Type,
                                                                             "=");
   function Find (Directory_Entries : Directory_Entry_Vectors.Vector;
                  Tag : Tag_Enum) return Directory_Entry_Type;
   Tag_Not_Found : exception;

   type Sample_Format_Enum is (Unsigned_Integer,
                               Signed_Integer,
                               Ieee_Float);
   Unknown_Sample_Format : exception;
   type Data_Width_Enum is (Width_8, Width_32);
   Unsupported_Data_Width : exception;

   subtype Unsigned_Integer_8_Array is Byte_Array;
   subtype Unsigned_Integer_32_Array is Long_Array;
   subtype Signed_Integer_32_Array is Slong_Array;
   subtype Ieee_Float_Array is Float_Array;

   type Row_Data_Type (Sample_Format_Value : Sample_Format_Enum;
                       Nb_Elements : Positive;
                       Data_Width : Data_Width_Enum) is record
      case Sample_Format_Value is
         when Unsigned_Integer =>
            case Data_Width is
               when Width_8 =>
                  Unsigned_8_Data : Unsigned_Integer_8_Array (1 .. Nb_Elements);
               when Width_32 =>
                  Unsigned_32_Data : Unsigned_Integer_32_Array (1 .. Nb_Elements);
            end case;
         when Signed_Integer =>
            Signed_32_Data : Signed_Integer_32_Array (1 .. Nb_Elements);
         when Ieee_Float =>
            Ieee_Float_Data : Ieee_Float_Array (1 .. Nb_Elements);
      end case;
   end record;

   package Row_Vectors is new Ada.Containers.Indefinite_Vectors (Positive,
                                                                 Row_Data_Type,
                                                                 "=");

   type Image_Kind_Enum is (Bilevel, Grayscale, RGB_Full_Color, Palette_Color);

   type Image_Type is record
      Header : Image_Header_Type;
      Kind : Image_Kind_Enum;
      Directories : Directory_Entry_Vectors.Vector;
      Rows : Row_Vectors.Vector;
   end record;

   One_Kilo_Byte : constant := 1024;
   One_Mega_Byte : constant := 1024 * One_Kilo_Byte;
   One_Giga_Byte : constant := 1024 * One_Mega_Byte;

   function Image (File_Name : String; Memory_Footprint_Limit : Long_Integer := One_Giga_Byte) return Image_Type;

end P42.Images.Tiff;
