with Ada.Numerics.Generic_Elementary_Functions;

package body P42.Images.Tools is

   package IEEE_Float_Elementary_Functions is new Ada.Numerics.Generic_Elementary_Functions (Interfaces.IEEE_Float_32);

   function Get_Min_Max (Matrix : Float_Matrix) return Float_Array is
      Min_Max : Float_Array (1 .. 2) :=
        (Interfaces.IEEE_Float_32'Last, Interfaces.IEEE_Float_32'First);
   begin
      for J in Matrix'Range (2) loop
         for I in Matrix'Range (1) loop
            if Matrix (I, J) < Min_Max (1) then
               Min_Max (1) := Matrix (I, J);
            end if;
            if Matrix (I, J) > Min_Max (2) then
               Min_Max (2) := Matrix (I, J);
            end if;
         end loop;
      end loop;
      return Min_Max;
   end Get_Min_Max;

   function Get_Distribution (Matrix : Float_Matrix;
                              Min : Interfaces.IEEE_Float_32;
                              Max : Interfaces.IEEE_Float_32;
                              Nb_Bin : Positive;
                              Exclude_Min : Boolean := False) return Longlong_Array is
      The_Step : constant Interfaces.IEEE_Float_32 := Step (Min, Max, Nb_Bin);
      Distribution : Longlong_Array (1 .. Nb_Bin) := (others => 0);
      Position : Positive;
      use type Interfaces.Unsigned_64;
   begin
      for J in Matrix'Range (2) loop
         for I in Matrix'Range (1) loop
            if Exclude_Min then
               if Matrix (I, J) > Min and Matrix (I, J) < Max then
                  Position := 1 + Natural ((Matrix (I, J) - Min) / The_Step);
                  Distribution (Position) := Distribution (Position) + 1;
               elsif Matrix (I, J) = Max then
                  Distribution (Distribution'Last) := Distribution (Distribution'Last) + 1;
               end if;
            else
               if Matrix (I, J) >= Min and Matrix (I, J) < Max then
                  Position := 1 + Natural ((Matrix (I, J) - Min) / The_Step);
                  Distribution (Position) := Distribution (Position) + 1;
               elsif Matrix (I, J) = Max then
                  Distribution (Distribution'Last) := Distribution (Distribution'Last) + 1;
               end if;
            end if;
         end loop;
      end loop;
      return Distribution;
   end Get_Distribution;

   procedure Split_With_Interval (Base : in out Fits.Image_Type;
                                  Depth : Positive;
                                  Min : Interfaces.IEEE_Float_32;
                                  Max : Interfaces.IEEE_Float_32;
                                  Default_Value : Interfaces.IEEE_Float_32) is
      New_Data : Fits.Data_Row_Vectors.Vector;
      Data : constant Fits.Data_Row_Vectors.Vector := Base.Datas.Element (Depth);
   begin
      for Row of Data loop
         declare
            New_Row : Fits.Data_Row_Type (Kind => Float, Length => Row.Length);
         begin
            for I in Row.Floats'Range loop
               if Row.Floats (I) >= Min and Row.Floats (I) < Max then
                  New_Row.Floats (I) := Row.Floats (I);
               else
                  New_Row.Floats (I) := Default_Value;
               end if;
            end loop;
            New_Data.Append (New_Row);
         end;
      end loop;
      Fits.Append (Base, New_Data, Min'Img & " split" & Max'Img);
   end Split_With_Interval;

   procedure Propagate_To_Neighbours (Matrix : in out Float_Matrix;
                                      Read : Float_Matrix;
                                      I : Positive;
                                      J : Positive;
                                      Cluster_Number : Interfaces.IEEE_Float_32;
                                      Background : Interfaces.IEEE_Float_32);

   function Identify_Clusters (Base : Fits.Image_Type;
                               Depth : Positive;
                               Background : Interfaces.IEEE_Float_32) return Fits.Data_Matrix_Type is
      Cluster_Matrix : Fits.Data_Matrix_Type (Kind => Float,
                                              Length1 => Base.Naxis1,
                                              Length2 => Base.Naxis2);
      Read_Matrix : constant Fits.Data_Matrix_Type := Fits.To_Matrix (Base.Datas.Element (Depth));
      Cluster_Number : Interfaces.IEEE_Float_32 := 1.0;
   begin
      Cluster_Matrix.Floats := (others => (others => 0.0));
      for J in Cluster_Matrix.Floats'Range (2) loop
         for I in Cluster_Matrix.Floats'Range (1) loop
            if Read_Matrix.Floats (I, J) > Background and Cluster_Matrix.Floats (I, J) = 0.0 then
               Propagate_To_Neighbours (Cluster_Matrix.Floats,
                                        Read_Matrix.Floats,
                                        I,
                                        J,
                                        Cluster_Number,
                                        Background);
               Cluster_Number := Cluster_Number + 1.0;
            end if;
         end loop;
      end loop;
      return Cluster_Matrix;
   end Identify_Clusters;

   procedure Propagate_To_Neighbours (Matrix : in out Float_Matrix;
                                      Read : Float_Matrix;
                                      I : Positive;
                                      J : Positive;
                                      Cluster_Number : Interfaces.IEEE_Float_32;
                                      Background : Interfaces.IEEE_Float_32) is
   begin
      if Matrix (I, J) /= 0.0 then
         if Matrix (I, J) /= Cluster_Number then
            raise Program_Error with Matrix (I, J)'Img & Cluster_Number'Img &
              " two clusters side by side";
         end if;
         return;
      end if;
      if Read (I, J) > Background then
         Matrix (I, J) := Cluster_Number;
         for Y in -1 .. 1 loop
            for X in -1 .. 1 loop
               if I + X in Matrix'Range (1) then
                  if J + Y in Matrix'Range (2) then
                     Propagate_To_Neighbours (Matrix, Read, I + X, J + Y, Cluster_Number, Background);
                  end if;
               end if;
            end loop;
         end loop;
      end if;
   end Propagate_To_Neighbours;

   function Get_Masses (Matrix : Float_Matrix;
                        Max_Cluster_Number : Positive) return Longlong_Array is
      Masses : Longlong_Array (1 .. Max_Cluster_Number) := (others => 0);
      use type Interfaces.Unsigned_64;
   begin
      for J in Matrix'Range (2) loop
         for I in Matrix'Range (1) loop
            if Matrix (I, J) <= Interfaces.IEEE_Float_32 (Max_Cluster_Number) and
              Matrix (I, J) /= 0.0
            then
               Masses (Positive (Matrix (I, J))) := Masses (Positive (Matrix (I, J))) + 1;
            end if;
         end loop;
      end loop;
      return Masses;
   end Get_Masses;

   function Get_Statistics (Matrix : Float_Matrix;
                            Masses : Longlong_Array) return Float_Matrix is
      Statistics : Float_Matrix (1 .. Masses'Length, 1 .. 5) := (others => (others => 0.0));
   begin
      for J in Matrix'Range (2) loop
         for I in Matrix'Range (1) loop
            if Matrix (I, J) /= 0.0 then
               declare
                  Cluster_Index : constant Positive := Positive (Matrix (I, J));
                  X : constant Interfaces.IEEE_Float_32 := Interfaces.IEEE_Float_32 (I);
                  Y : constant Interfaces.IEEE_Float_32 := Interfaces.IEEE_Float_32 (J);
               begin
                  Statistics (Cluster_Index, 1) := Statistics (Cluster_Index, 1) + X;
                  Statistics (Cluster_Index, 2) := Statistics (Cluster_Index, 2) + Y;
               end;
            end if;
         end loop;
      end loop;
      for I in Masses'Range loop
         Statistics (I, 1) := Statistics (I, 1) / Interfaces.IEEE_Float_32 (Masses (I));
         Statistics (I, 2) := Statistics (I, 2) / Interfaces.IEEE_Float_32 (Masses (I));
      end loop;
      for J in Matrix'Range (2) loop
         for I in Matrix'Range (1) loop
            if Matrix (I, J) /= 0.0 then
               declare
                  Cluster_Index : constant Positive := Positive (Matrix (I, J));
                  X : constant Interfaces.IEEE_Float_32 := Interfaces.IEEE_Float_32 (I);
                  Y : constant Interfaces.IEEE_Float_32 := Interfaces.IEEE_Float_32 (J);
                  Diff_X : constant Interfaces.IEEE_Float_32 := X - Statistics (Cluster_Index, 1);
                  Diff_X_2 : constant Interfaces.IEEE_Float_32 := Diff_X * Diff_X;
                  Diff_Y : constant Interfaces.IEEE_Float_32 := Y - Statistics (Cluster_Index, 2);
                  Diff_Y_2 : constant Interfaces.IEEE_Float_32 := Diff_Y * Diff_Y;
               begin
                  Statistics (Cluster_Index, 3) := Statistics (Cluster_Index, 3) + Diff_X_2;
                  Statistics (Cluster_Index, 4) := Statistics (Cluster_Index, 4) + Diff_Y_2;
                  Statistics (Cluster_Index, 5) := Statistics (Cluster_Index, 5) + Diff_X_2 + Diff_Y_2;
               end;
            end if;
         end loop;
      end loop;
      for I in Masses'Range loop
         Statistics (I, 3) := Statistics (I, 3) / Interfaces.IEEE_Float_32 (Masses (I));
         Statistics (I, 4) := Statistics (I, 4) / Interfaces.IEEE_Float_32 (Masses (I));
         Statistics (I, 5) := Statistics (I, 5) / Interfaces.IEEE_Float_32 (Masses (I));
         Statistics (I, 3) := IEEE_Float_Elementary_Functions.Sqrt (Statistics (I, 3));
         Statistics (I, 4) := IEEE_Float_Elementary_Functions.Sqrt (Statistics (I, 4));
         Statistics (I, 5) := IEEE_Float_Elementary_Functions.Sqrt (Statistics (I, 5));
      end loop;
      return Statistics;
   end Get_Statistics;

   function Compress_Distribution (Distribution : Longlong_Array;
                                   Compression_Factor : Positive) return Longlong_Array is
      Is_Round : constant Boolean := (Distribution'Length mod Compression_Factor) = 0;
      Distribution_Index : Positive := Distribution'First;
      Base_Index : constant Positive := Distribution'Length / Compression_Factor;
      Last_Index : constant Positive := (if Is_Round then Base_Index else Base_Index + 1);
      New_Distribution : Longlong_Array (1 .. Last_Index) := (others => 0);
   begin
      for I in New_Distribution'Range loop
         for J in 0 .. Compression_Factor - 1 loop
            New_Distribution (I) := Distribution (Distribution_Index);
            Distribution_Index := Distribution_Index + 1;
            exit when Distribution_Index > Distribution'Last;
         end loop;
      end loop;
      return New_Distribution;
   end Compress_Distribution;

   function Calcul_Distribution (Data : Longlong_Array) return Longlong_Array is
      Max_Value : Interfaces.Unsigned_64 := 0;
      use type Interfaces.Unsigned_64;
   begin
      for I in Data'Range loop
         if Data (I) > Max_Value then
            Max_Value := Data (I);
         end if;
      end loop;
      declare
         Distribution : Longlong_Array (1 .. Positive (Max_Value)) := (others => 0);
      begin
         for I in Data'Range loop
            if Data (I) /= 0 then
               Distribution (Positive (Data (I))) := Distribution (Positive (Data (I))) + 1;
            end if;
         end loop;
         return Distribution;
      end;
   end Calcul_Distribution;

end P42.Images.Tools;
