with Ada.Streams;
with System;

with POSIX.Calendar;
with POSIX.Implementation;

with P42.Streams;

package body P42.Messages_Queues is

   use type Ada.Streams.Stream_Element_Offset;

   procedure Send (Queue : POSIX.Message_Queues.Message_Queue_Descriptor;
                   Message : Message_Type) is
      Stream : aliased Streams.Streamer_Type (Message_Type'Size / 8);
   begin
      Stream.Initialise;
      Message_Type'Output (Stream'Access, Message);
      POSIX.Message_Queues.Send (Queue,
                                 Stream.Data,
                                 Default_Priority);
   end Send;

   function Receive (Queue : POSIX.Message_Queues.Message_Queue_Descriptor) return Message_Type is
      Stream : aliased Streams.Streamer_Type (Message_Type'Size / 8);
      Priority : POSIX.Message_Queues.Message_Priority;
   begin
      POSIX.Message_Queues.Receive (Queue, Stream.Data, Stream.Last, Priority);
      return Message_Type'Input (Stream'Access);
   end Receive;

   mq_timedreceive_LINKNAME : constant String := "mq_timedreceive";

   function mq_timedreceive
     (mqdes    : POSIX.Message_Queues.Message_Queue_Descriptor;
      msg_ptr  : System.Address;
      msg_len  : POSIX.C.size_t;
      msg_prio : access POSIX.C.unsigned;
      msg_timeout : access POSIX.C.struct_timespec) return POSIX.C.ssize_t;
   pragma Import (C, mq_timedreceive, mq_timedreceive_LINKNAME);

   procedure Posix_Timed_Receive
     (MQ             : POSIX.Message_Queues.Message_Queue_Descriptor;
      Message        : out Ada.Streams.Stream_Element_Array;
      Last           : out Ada.Streams.Stream_Element_Offset;
      Priority       : out POSIX.Message_Queues.Message_Priority;
      Abs_Timeout    : POSIX.C.struct_timespec;
      Masked_Signals : POSIX.Signal_Masking := POSIX.RTS_Signals);

   procedure Posix_Timed_Receive
     (MQ             : POSIX.Message_Queues.Message_Queue_Descriptor;
      Message        : out Ada.Streams.Stream_Element_Array;
      Last           : out Ada.Streams.Stream_Element_Offset;
      Priority       : out POSIX.Message_Queues.Message_Priority;
      Abs_Timeout    : POSIX.C.struct_timespec;
      Masked_Signals : POSIX.Signal_Masking := POSIX.RTS_Signals) is
      Old_Mask : aliased POSIX.Implementation.Signal_Mask;
      Prio : aliased POSIX.C.unsigned;
      Timeout : aliased POSIX.C.struct_timespec := Abs_Timeout;
      Result : POSIX.C.ssize_t;
      use POSIX.Message_Queues;
   begin
      POSIX.Implementation.Mask_Signals (Masked_Signals,
                                         Old_Mask'Unchecked_Access);
      Result := mq_timedreceive
        (MQ,
         Message (Message'First)'Address,
         POSIX.C.size_t (Message'Length),
         Prio'Unchecked_Access,
         Timeout'Unchecked_Access);
      POSIX.Implementation.Check_NNeg_And_Restore_Signals
        (POSIX.C.int (Result), Masked_Signals, Old_Mask'Unchecked_Access);
      Priority := Message_Priority (Prio);
      Last := Message'First + Ada.Streams.Stream_Element_Offset (Result) - 1;
   end Posix_Timed_Receive;

   function Timed_Receive
     (Queue : POSIX.Message_Queues.Message_Queue_Descriptor;
      Abs_Timeout : POSIX.C.struct_timespec)
     return Message_Type is
      Priority : POSIX.Message_Queues.Message_Priority;
      Stream : aliased Streams.Streamer_Type (Message_Type'Size / 8);
   begin
      Posix_Timed_Receive (Queue,
                           Stream.Data,
                           Stream.Last,
                           Priority,
                           Abs_Timeout);
      return Message_Type'Input (Stream'Access);
   end Timed_Receive;

   function Relative_Timed_Receive
     (Queue : POSIX.Message_Queues.Message_Queue_Descriptor;
      Relative_Timeout : Natural)
     return Message_Type is
      Priority : POSIX.Message_Queues.Message_Priority;
      Abs_Timeout : POSIX.C.struct_timespec;
      Current_Time : constant POSIX.Calendar.POSIX_Time :=
        POSIX.Calendar.Clock;
      Current_Time_Raw : constant POSIX.C.struct_timespec :=
        POSIX.Implementation.To_Struct_Timespec
        (POSIX.Calendar.To_Timespec (Current_Time));
      use type POSIX.C.long;
      use type POSIX.C.time_t;
      Stream : aliased Streams.Streamer_Type (Message_Type'Size / 8);
   begin
      Abs_Timeout.tv_nsec := Current_Time_Raw.tv_nsec;
      Abs_Timeout.tv_sec := POSIX.C.time_t (Relative_Timeout) +
        Current_Time_Raw.tv_sec;
      Posix_Timed_Receive (Queue,
                           Stream.Data,
                           Stream.Last,
                           Priority,
                           Abs_Timeout);
      return Message_Type'Input (Stream'Access);
   end Relative_Timed_Receive;

   function Relative_Timed_Receive
     (Queue : POSIX.Message_Queues.Message_Queue_Descriptor;
      Relative_Timeout : Duration)
     return Message_Type is
      Priority : POSIX.Message_Queues.Message_Priority;
      Current_Time : constant POSIX.Calendar.POSIX_Time :=
        POSIX.Calendar.Clock;
      use type POSIX.Calendar.POSIX_Time;
      Timeout_Time : constant POSIX.Calendar.POSIX_Time :=
        Relative_Timeout + Current_Time;
      Abs_Timeout : constant POSIX.C.struct_timespec :=
        POSIX.Implementation.To_Struct_Timespec
        (POSIX.Calendar.To_Timespec (Timeout_Time));
      Stream : aliased Streams.Streamer_Type (Message_Type'Size / 8);
   begin
      Posix_Timed_Receive (Queue,
                           Stream.Data,
                           Stream.Last,
                           Priority,
                           Abs_Timeout);
      return Message_Type'Input (Stream'Access);
   end Relative_Timed_Receive;


end P42.Messages_Queues;
