with Ada.Characters.Handling;
with Ada.Directories;
with Ada.Streams.Stream_IO;
with Ada.Strings.Fixed;
with Ada.Strings.Unbounded;
with Ada.Text_IO;

package body P42.ShaXXX is

   Block_Length : constant Natural := Unsigned_Type'Size * 2;
   Block_Length_2 : constant Natural := Unsigned_Type'Size * 4;
   Block_Offset : constant Stream_Element_Offset := Unsigned_Type'Size * 2;

   function Get_Unsigned_Type (Message : Stream_Element_Array;
                               Offset : Stream_Element_Offset) return Unsigned_Type
     with Pre => (Offset - Message'First) mod (Unsigned_Type'Size / 8) = 0;

   type Huge_Unsigned_Type is record
      Hi : Unsigned_Type;
      Lo : Unsigned_Type;
   end record;
   function "+" (Left : Huge_Unsigned_Type; Right : Unsigned_Type) return Huge_Unsigned_Type;
   function "-" (Left : Unsigned_Type; Right : Huge_Unsigned_Type) return Huge_Unsigned_Type;
   function Shift_Left (Value : Huge_Unsigned_Type; Amount : Natural) return Huge_Unsigned_Type;
   function Shift_Right (Value : Huge_Unsigned_Type; Amount : Natural) return Huge_Unsigned_Type;
   function Huge_To_Stream (Value : Huge_Unsigned_Type) return Stream_Element_Array;
   function Modulus (Value : Huge_Unsigned_Type; Modulo : Unsigned_Type) return Huge_Unsigned_Type;

   function To_Stream (Value : Unsigned_Type) return Stream_Element_Array;

   package Unsigned_IO is new Ada.Text_IO.Modular_IO (Unsigned_Type);

   function "+" (Left : Huge_Unsigned_Type; Right : Unsigned_Type) return Huge_Unsigned_Type is
      Result : Huge_Unsigned_Type;
   begin
      Result.Lo := Left.Lo + Right;
      Result.Hi := Left.Hi;
      if Result.Lo < Right then
         Result.Hi := Result.Hi + 1;
      end if;
      return Result;
   end "+";

   function "-" (Left : Unsigned_Type; Right : Huge_Unsigned_Type) return Huge_Unsigned_Type is
      Result : Huge_Unsigned_Type;
   begin
      Result.Lo := Left - Right.Lo;
      Result.Hi := Right.Hi;
      if Result.Lo > Left then
         Result.Hi := Result.Hi - 1;
      end if;
      return Result;
   end "-";

   function Shift_Left (Value : Huge_Unsigned_Type; Amount : Natural) return Huge_Unsigned_Type is
      Result : Huge_Unsigned_Type;
      Mask : Unsigned_Type := 0;
   begin
      if Amount = 0 then
         return Value;
      end if;
      for I in 1 .. Amount loop
         Mask := Mask or Shift_Left (2#1#, Unsigned_Type'Size - Amount);
      end loop;
      Result.Hi := Shift_Left (Value.Hi, Amount);
      Result.Hi := Result.Hi or Shift_Right (Value.Lo and Mask, Unsigned_Type'Size - Amount);
      Result.Lo := Shift_Left (Value.Lo, Amount);
      return Result;
   end Shift_Left;

   function Shift_Right (Value : Huge_Unsigned_Type; Amount : Natural) return Huge_Unsigned_Type is
      Result : Huge_Unsigned_Type;
      Mask : Unsigned_Type := 0;
   begin
      if Amount = 0 then
         return Value;
      end if;
      Mask := 1;
      for I in 1 .. Amount - 1 loop
         Mask := Mask or Shift_Left (2#1#, Amount);
      end loop;
      Result.Hi := Shift_Right (Value.Hi, Amount);
      Result.Lo := Shift_Right (Value.Lo, Amount);
      Result.Lo := Result.Lo or Shift_Left (Value.Hi and Mask, Unsigned_Type'Size - Amount);
      return Result;
   end Shift_Right;

   function Huge_To_Stream (Value : Huge_Unsigned_Type) return Stream_Element_Array is
      Data : Stream_Element_Array (1 .. Huge_Unsigned_Type'Size / 8);
   begin
      Data (1 .. Unsigned_Type'Size / 8) := To_Stream (Value.Hi);
      Data (1 + Unsigned_Type'Size / 8 .. Data'Last) := To_Stream (Value.Lo);
      return Data;
   end Huge_To_Stream;

   function Modulus (Value : Huge_Unsigned_Type; Modulo : Unsigned_Type) return Huge_Unsigned_Type is
      Result : Huge_Unsigned_Type;
   begin
      Result.Hi := 0;
      Result.Lo := Value.Lo mod Modulo;
      return Result;
   end Modulus;

   function To_Stream (Value : Unsigned_Type) return Stream_Element_Array is
      New_Value : Unsigned_Type;
      Data : Stream_Element_Array (1 .. Unsigned_Type'Size / 8);
      for Data'Address use New_Value'Address;
   begin
      New_Value := To_Big_Endian (Value);
      return Data;
   end To_Stream;

   Null_Array : Stream_Element_Array (1 .. 0);

   function Padding (Message : Stream_Element_Array;
                     Length : Huge_Unsigned_Type) return Stream_Element_Array
     with Post => Padding'Result'Length = Block_Length or Padding'Result'Length = Block_Length_2;

   -------------
   -- Padding --
   -------------

   function Padding
     (Message : Stream_Element_Array;
      Length : Huge_Unsigned_Type)
      return Stream_Element_Array
   is
      K : Huge_Unsigned_Type;
      K_Bis : Huge_Unsigned_Type;
      Length_Decaled : constant Huge_Unsigned_Type := Shift_Left (Length, 3);
   begin
      K := Modulus ((Ref_Padding - (Length_Decaled + 1)), Mod_Padding);
      K_Bis := Shift_Right (K + 1, 3);
      declare
         Data_Length : constant Stream_Element_Array := Huge_To_Stream (Length_Decaled);
         New_Message : Stream_Element_Array
           (Message'First .. Message'Last + Data_Length'Length + Stream_Element_Offset (K_Bis.Lo));
      begin
         New_Message (Message'First .. Message'Last) := Message;
         New_Message (Message'Last + 1) := 2#1000_0000#;
         if Message'Last + 2 = New_Message'Last + 1 - Huge_Unsigned_Type'Size / 8 then
            New_Message (Message'Last + 2 .. New_Message'Last) := Data_Length;
         else
            for I in Message'Last + 2 .. New_Message'Last - Huge_Unsigned_Type'Size / 8 loop
               New_Message (I) := 0;
            end loop;
            New_Message (New_Message'Last + 1 - Huge_Unsigned_Type'Size / 8 .. New_Message'Last) := Data_Length;
         end if;
         return New_Message;
      end;
   end Padding;

   procedure Hash_Computation_Block (Context : in out Context_Type;
                                     Block : Stream_Element_Array)
     with Pre => Block'Length = Block_Length;

   procedure Hash_Computation_Blocks (Context : in out Context_Type;
                                      Block : Stream_Element_Array)
     with Pre => Block'Length mod Block_Length = 0;

   procedure End_Hash_Computation (Context : in out Context_Type;
                                   Block : Stream_Element_Array;
                                   Length : Huge_Unsigned_Type);

   procedure Hash_Computation (Context : out Context_Type;
                               Message : Stream_Element_Array) is
   begin
      Context.Ctx := H_Init;
      End_Hash_Computation (Context, Message, (Hi => 0, Lo => Message'Length));
   end Hash_Computation;

   procedure End_Hash_Computation (Context : in out Context_Type;
                                   Block : Stream_Element_Array;
                                   Length : Huge_Unsigned_Type) is
      Number_Of_Blocks : constant Stream_Element_Offset := Block'Length / Block_Offset;
      Padding_On_Boundary : constant Boolean := Block'Length mod Block_Length = 0;
   begin
      if Number_Of_Blocks /= 0 then
         Hash_Computation_Blocks (Context, Block (Block'First .. Block'First + Number_Of_Blocks * Block_Offset - 1));
      end if;
      if Padding_On_Boundary then
         declare
            End_Message : constant Stream_Element_Array := Padding (Null_Array, Length);
         begin
            Hash_Computation_Block (Context, End_Message);
         end;
      else
         declare
            End_Message : constant Stream_Element_Array :=
              Padding (Block (Block'First + Number_Of_Blocks * Block_Offset .. Block'Last), Length);
         begin
            Hash_Computation_Blocks (Context, End_Message);
         end;
      end if;
   end End_Hash_Computation;

   procedure Hash_Computation (Context : out Context_Type;
                               File_Name : String) is
      File : Stream_IO.File_Type;
      Buffer : Buffer_Array;
      Last : Stream_Element_Offset;
      Length : Huge_Unsigned_Type := (0, 0);
      use type Ada.Directories.File_Size;
   begin
      Stream_IO.Open (File, Stream_IO.In_File, File_Name);
      if Ada.Directories.Size (File_Name) < Buffer_Array'Length then
         Stream_IO.Read (File, Buffer, Last);
         Hash_Computation (Context, Buffer (Buffer'First .. Last));
      else
         Context.Ctx := H_Init;
         loop
            exit when Stream_IO.End_Of_File (File);
            Stream_IO.Read (File, Buffer, Last);
            Length := Length + Unsigned_Type (Last);
            if Last = Buffer'Last then
               Hash_Computation_Blocks (Context, Buffer);
               if Stream_IO.End_Of_File (File) then
                  declare
                     End_Message : constant Stream_Element_Array := Padding (Null_Array,
                                                                             Length);
                  begin
                     Hash_Computation_Block (Context, End_Message);
                     exit;
                  end;
               end if;
            else
               End_Hash_Computation (Context, Buffer (Buffer'First .. Last), Length);
               exit;
            end if;
         end loop;
      end if;
      Stream_IO.Close (File);
   end Hash_Computation;

   procedure Hash_Computation_Block (Context : in out Context_Type;
                                     Block : Stream_Element_Array) is
      W : Unsigned_Array (K'First .. K'Last);
      A, B, C, D, E, F, G, H, T_1, T_2 : Unsigned_Type;
      Ctx : Unsigned_Array renames Context.Ctx;
   begin
      for T in 0 .. 15 loop
         W (T) := To_Big_Endian
           (Get_Unsigned_Type (Block,
                               Block'First + Stream_Element_Offset (T) * Unsigned_Type'Size / 8));
      end loop;
      for T in 16 .. K'Last loop
         W (T) := Sigma_1 (W (T - 2)) + W (T - 7) + Sigma_0 (W (T - 15)) + W (T - 16);
      end loop;
      A := Ctx (0);
      B := Ctx (1);
      C := Ctx (2);
      D := Ctx (3);
      E := Ctx (4);
      F := Ctx (5);
      G := Ctx (6);
      H := Ctx (7);
      for T in K'Range loop
         T_1 := H + Sum_1 (E) + Ch (E, F, G) + K (T) + W (T);
         T_2 := Sum_0 (A) + Maj (A, B, C);
         H := G;
         G := F;
         F := E;
         E := D + T_1;
         D := C;
         C := B;
         B := A;
         A := T_1 + T_2;
      end loop;
      Ctx (0) := Ctx (0) + A;
      Ctx (1) := Ctx (1) + B;
      Ctx (2) := Ctx (2) + C;
      Ctx (3) := Ctx (3) + D;
      Ctx (4) := Ctx (4) + E;
      Ctx (5) := Ctx (5) + F;
      Ctx (6) := Ctx (6) + G;
      Ctx (7) := Ctx (7) + H;
   end Hash_Computation_Block;

   procedure Hash_Computation_Blocks (Context : in out Context_Type;
                                      Block : Stream_Element_Array) is
      Number_Of_Blocks : constant Stream_Element_Offset := Block'Length / Block_Offset;
   begin
      for I in 1 .. Number_Of_Blocks loop
         Hash_Computation_Block
           (Context,
            Block (Block'First + (I - 1) * Block_Offset .. Block'First + I * Block_Offset - 1));
      end loop;
   end Hash_Computation_Blocks;

   function Get_Unsigned_Type (Message : Stream_Element_Array;
                               Offset : Stream_Element_Offset) return Unsigned_Type is
      pragma Warnings (Off);  --  Precondition on offset ensure Address is aligned
      Data : Unsigned_Type;
      for Data'Address use Message (Offset)'Address;
      pragma Warnings (On);
   begin
      return Data;
   end Get_Unsigned_Type;

   function Hash_Value_Image (Value : Unsigned_Type) return String
     with Post => Hash_Value_Image'Result'Length = Unsigned_Type'Size / 4;

   function Image (Context : Context_Type) return String is
      Ctx : Unsigned_Array renames Context.Ctx;
   begin
      if Cut_Output_Size  = 1 then
         return Hash_Value_Image (Ctx (Ctx'First)) &
           Hash_Value_Image (Ctx (Ctx'First + 1)) &
           Hash_Value_Image (Ctx (Ctx'First + 2)) &
           Hash_Value_Image (Ctx (Ctx'First + 3)) &
           Hash_Value_Image (Ctx (Ctx'First + 4)) &
           Hash_Value_Image (Ctx (Ctx'First + 5)) &
           Hash_Value_Image (Ctx (Ctx'First + 6));
      elsif Cut_Output_Size  = 2 then
         return Hash_Value_Image (Ctx (Ctx'First)) &
           Hash_Value_Image (Ctx (Ctx'First + 1)) &
           Hash_Value_Image (Ctx (Ctx'First + 2)) &
           Hash_Value_Image (Ctx (Ctx'First + 3)) &
           Hash_Value_Image (Ctx (Ctx'First + 4)) &
           Hash_Value_Image (Ctx (Ctx'First + 5));
      else
         return Hash_Value_Image (Ctx (Ctx'First)) &
           Hash_Value_Image (Ctx (Ctx'First + 1)) &
           Hash_Value_Image (Ctx (Ctx'First + 2)) &
           Hash_Value_Image (Ctx (Ctx'First + 3)) &
           Hash_Value_Image (Ctx (Ctx'First + 4)) &
           Hash_Value_Image (Ctx (Ctx'First + 5)) &
           Hash_Value_Image (Ctx (Ctx'First + 6)) &
           Hash_Value_Image (Ctx (Ctx'First + 7));
      end if;
   end Image;

   function Hash_Value_Image (Value : Unsigned_Type) return String is
      Local_String : String (1 .. 4 + Unsigned_Type'Size / 4);
      Return_Value : String (1 .. Unsigned_Type'Size / 4) := (others => '0');
      Base_Position : Natural := 0;
   begin
      Unsigned_IO.Put (Local_String, Value, 16);
      Base_Position := Ada.Strings.Fixed.Index (Local_String, "16#");
      for I in Base_Position + 3 .. 3 + Unsigned_Type'Size / 4 loop
         Return_Value (Return_Value'Length + I - (3 + Unsigned_Type'Size / 4)) := Local_String (I);
      end loop;
      return Ada.Characters.Handling.To_Lower (Return_Value);
   end Hash_Value_Image;

   task body Hash_Computation_Task is
      Our_Context : Context_Type;
      File : Ada.Strings.Unbounded.Unbounded_String;
      use Ada.Strings.Unbounded;
   begin
      loop
         select
            accept Start (File_Name : String) do
               File := To_Unbounded_String (File_Name);
            end Start;
         or
            terminate;
         end select;
         Hash_Computation (Our_Context, To_String (File));
         select
            accept Get_Result (Context : out Context_Type) do
               Context := Our_Context;
            end Get_Result;
         or
            terminate;
         end select;
      end loop;
   end Hash_Computation_Task;

end P42.ShaXXX;
