with Ada.Containers;
with Ada.Strings.Fixed;
with Ada.Strings.Unbounded;

with GNAT.Regpat;

package body P42.String_Manipulation is

   ----------------
   -- Substitute --
   ----------------

   function Substitute
     (Source : String;
      Origin : String;
      By : String;
      Recursive : Boolean := False)
      return String
   is
      Origin_Position : constant Natural := Ada.Strings.Fixed.Index (Source, Origin);
      Last_Position : constant Natural := Origin_Position + Origin'Length;
   begin
      if Origin_Position = 0 then
         return Source;
      elsif Origin_Position = Source'First and Last_Position - 1 = Source'Last then
         return By;
      elsif Origin_Position = Source'First then
         if Recursive then
            return Substitute (Source => By & Source (Last_Position .. Source'Last),
                               Origin => Origin,
                               By => By,
                               Recursive => True);
         else
            return By & Source (Last_Position .. Source'Last);
         end if;
      elsif Last_Position - 1 = Source'Last then
         return Source (Source'First .. Origin_Position - 1) & By;
      else
         declare
            Intermediary_Res : constant String :=
              Source (Source'First .. Origin_Position - 1) & By & Source (Last_Position .. Source'Last);
         begin
            if Recursive then
               return Substitute (Source => Intermediary_Res,
                                  Origin => Origin,
                                  By => By,
                                  Recursive => True);
            else
               return Intermediary_Res;
            end if;
         end;
      end if;
   end Substitute;

   function Substitute (Source : String;
                        Templates : P42.String_Couple_Vectors.Vector;
                        Recursive : Boolean := False) return String is
      use Ada.Strings.Unbounded;
      Temporary_String : Unbounded_String := To_Unbounded_String (Source);
   begin

      for Couple of Templates loop
         Temporary_String := To_Unbounded_String (Substitute (Source => To_String (Temporary_String),
                                                              Origin => Couple.Key,
                                                              By => Couple.Value,
                                                              Recursive => Recursive));
      end loop;
      return To_String (Temporary_String);
   end Substitute;

   function Substitute (Source : String;
                        Pattern : GNAT.Regpat.Pattern_Matcher;
                        By : String;
                        Recursive : Boolean) return String;
   function Substitute (Source : String;
                        Pattern : GNAT.Regpat.Pattern_Matcher;
                        By : String;
                        Recursive : Boolean) return String is
      use GNAT.Regpat;
      Matches : Match_Array (0 .. 0);
   begin
      Match (Pattern, Source, Matches);
      if Matches (0) = No_Match then
         return Source;
      end if;
      declare
         New_Source : constant String :=
           Source (Source'First .. Matches (0).First - 1) & By & Source (Matches (0).Last + 1 .. Source'Last);
      begin
         if Recursive then
            return Substitute (New_Source, Pattern, By, True);
         else
            return New_Source;
         end if;
      end;
   end Substitute;

   function Substitute_Regexp (Source : String;
                               Regexp : String;
                               By : String;
                               Recursive : Boolean := False) return String is
      Pattern : constant GNAT.Regpat.Pattern_Matcher := GNAT.Regpat.Compile (Regexp);
   begin
      return Substitute (Source, Pattern, By, Recursive);
   end Substitute_Regexp;

   function Remove_Character (Source : String;
                              C : Character;
                              Where : Position_Enum := Both) return String is
      First : Positive := Source'First;
      Last : Positive := Source'Last;
   begin
      if Source = "" then
         return "";
      end if;
      case Where is
         when Beginning =>
            loop
               exit when Source (First .. Last) = "";
               exit when Source (First) /= C;
               First := First + 1;
            end loop;
         when Ending =>
            loop
               exit when Source (First .. Last) = "";
               exit when Source (Last) /= C;
               Last := Last - 1;
            end loop;
         when Both =>
            return Remove_Spaces (Remove_Spaces (Source, Beginning), Ending);
      end case;
      return Source (First .. Last);
   end Remove_Character;

   function To_List_Indexed (Source : P42.String_Vectors.Vector;
                             Separator : String;
                             Index : Positive) return String;

   function To_List (Source : P42.String_Vectors.Vector;
                     Separator : String := " ") return String is
      use type Ada.Containers.Count_Type;
   begin
      if Source.Is_Empty then
         return "";
      else
         return To_List_Indexed (Source, Separator, Source.First_Index);
      end if;
   end To_List;

   function To_List_Indexed (Source : P42.String_Vectors.Vector;
                             Separator : String;
                             Index : Positive) return String is
   begin
      if Index = Source.Last_Index then
         return Source.Last_Element;
      else
         return Source.Element (Index) & Separator & To_List_Indexed (Source, Separator, Index + 1);
      end if;
   end To_List_Indexed;

end P42.String_Manipulation;
